/* This is the CORE CODE for the toolbox 'MARS', a faster version. It puts likelihoods computation out of C code, and uses Matlab to calculate likelihoods. */
/* It can accept three types of TCM. */
/* FORMAT deltaFEF = spm_mars(J,q,l,extE,lkp,rLabel); */
/*
 * Some codes are adapted from spm_mrf.c by John Ashburner
 *
 * (c) Yu Huang (Andy), May 2013, August 2014,
   The Neural Engineering Lab, Dept. of Biomedical Engineering, City College of New York
   Send bugs to yhuang16@citymail.cuny.edu
 *
 */
/*
 * See the following paper for details: Huang Y, Parra LC (2015) Fully Automated Whole-Head Segmentation with Improved Smoothness and Continuity, with Theory Reviewed. PLoS ONE 10(5): e0125477.
   doi:10.1371/journal.pone.0125477
 * This code implements the Eq. 2
 *
 */

#include <stdlib.h>
#include <math.h>
#include "matrix.h"
#include "mex.h"

#define TINY 1.0e-64
#define MAXCLASSES 30

static void neighborProcessGlobal(double interEng[], float J[], double qq[], int lkp[], mwSize dimStd[], mwSize dimFull[], int volSize)
{
int k, n;
/*mwSize offsetLocal;*/
double qqs[MAXCLASSES];
float *J1;

for(k=0; k<dimStd[3]; k++)
   qqs[k] = 0.0; /*initialize qqs*/

for(k=0; k<dimFull[3]; k++)
   qqs[lkp[k]] += qq[k*volSize]; /*shrink the multi-gaussian for each class*/

for(k=0; k<dimStd[3]; k++)
   { J1 = J + k;
     for(n=0; n<dimStd[3]; n++, J1+=dimStd[3]) /*dot product goes along each ROW of TCM matrix*/
       { interEng[k] += (*J1)*qqs[n];
         /*mexPrintf("code is %d, and current val in J is %f.\n",code,*J1);*/
       }
   }

}

static void neighborProcessLocal(double interEng[], float J[], double qq[], int lkp[], mwSize dimStd[], mwSize dimFull[], int volSize, mwSize offset, mwSize offsetLocal, mwSize offsetJ1)
{
int k, n;
/*mwSize offsetLocal;*/
double qqs[MAXCLASSES];
float *J1;

for(k=0; k<dimStd[3]; k++)
   qqs[k] = 0.0; /*initialize qqs*/

for(k=0; k<dimFull[3]; k++)
   qqs[lkp[k]] += qq[k*volSize]; /*shrink the multi-gaussian for each class*/

for(k=0; k<dimStd[3]; k++)
   { J1 = J + offset + k*volSize + offsetLocal; /*use neighborCode to encode which neighbor, so local TCM should have corresponding order, the last dim of local TCM should store K x K matrix for inferior, superior, posterior, anterior, left and right neighbor, from 1 to 6, respectively.*/
    for(n=0; n<dimStd[3]; n++, J1+=offsetJ1) /*dot product goes along each ROW of TCM matrix*/
      { interEng[k] += (*J1)*qqs[n];
        /*mexPrintf("code is %d, and current val in J is %f.\n",code,*J1);*/
      }
   }
}

static void neighborProcessRegional(double interEng[], float J[], double qq[], int lkp[], mwSize dimStd[], mwSize dimFull[], int volSize, mwSize offsetJ)
{
int k, n;
/*mwSize offsetLocal;*/
double qqs[MAXCLASSES];
float *J1;

for(k=0; k<dimStd[3]; k++)
   qqs[k] = 0.0; /*initialize qqs*/

for(k=0; k<dimFull[3]; k++)
   qqs[lkp[k]] += qq[k*volSize]; /*shrink the multi-gaussian for each class*/

for(k=0; k<dimStd[3]; k++)
   { J1 = J + offsetJ + k;
      for(n=0; n<dimStd[3]; n++, J1+=dimStd[3]) /*dot product goes along each ROW of TCM matrix*/
        { interEng[k] += (*J1)*qqs[n];
          /*mexPrintf("code is %d, and current val in J is %f.\n",code,*J1);*/
        }
   }

}

static double updateGlobal(float J[], double q[], double l[], double extEng[], int lkp[], mwSize dimStd[], mwSize dimFull[])
{
int it, k, volSize;
double Fold, Fnew, sqtmp, dF=0.0;
mwSize i0, i1, i2, i0start, i1start, i2start, offset;
double *q0, *l0, *extEng0;
double interEng[MAXCLASSES], ls[MAXCLASSES], qs[MAXCLASSES], qtmp[MAXCLASSES], tmp[MAXCLASSES];

volSize = dimStd[0]*dimStd[1]*dimStd[2];

for(it=0; it<2; it++)
    {
        i2start = it%2;
        for(i2=0; i2<dimStd[2]; i2++) /* Inferior -> Superior */
        {
            i1start = (i2start == (i2%2));
            for(i1=0; i1<dimStd[1]; i1++) /* Posterior -> Anterior */
            {
                i0start = (i1start == (i1%2));

                for(i0=i0start; i0<dimStd[0]; i0+=2) /* Left -> Right */
                {

                    offset = i0 + dimStd[0]*(i1 + dimStd[1]*i2);
/* NOTE: convert Matlab indexing into linear indexing, for COLUMN-first indexing (Matlab matrix style),
we have A(i0,i1,i2), is equivalent to A[i0 + dimStd[0]*(i1 + dimStd[1]*i2)], 0-based */

                    /* Pointers to current voxel in first volume */
                    q0 = q + offset;
                    l0 = l + offset;
                    extEng0 = extEng + offset;

                    for (k=0; k<dimStd[3]; k++)
                        { interEng[k] = 0.0;
                          ls[k] = 0.0;
                          qs[k] = 0.0; }  /*initialization*/

                    if (i2>0) /*inferior neighbor*/
                    neighborProcessGlobal(interEng,J,q0-dimStd[0]*dimStd[1],lkp,dimStd,dimFull,volSize);
                    if (i2<dimStd[2]-1) /*superior neighbor*/
                    neighborProcessGlobal(interEng,J,q0+dimStd[0]*dimStd[1],lkp,dimStd,dimFull,volSize);
                    if (i1>0) /*posterior neighbor*/
                    neighborProcessGlobal(interEng,J,q0-dimStd[0],lkp,dimStd,dimFull,volSize);
                    if (i1<dimStd[1]-1) /*anterior neighbor*/
                    neighborProcessGlobal(interEng,J,q0+dimStd[0],lkp,dimStd,dimFull,volSize);
                    if (i0>0) /*left neighbor*/
                    neighborProcessGlobal(interEng,J,q0-1,lkp,dimStd,dimFull,volSize);
                    if (i0<dimStd[0]-1) /*right neighbor*/
                    neighborProcessGlobal(interEng,J,q0+1,lkp,dimStd,dimFull,volSize);

                    for(k=0; k<dimFull[3]; k++)
                       { ls[lkp[k]] += l0[k*volSize];
                         qs[lkp[k]] += q0[k*volSize]; }  /*shrink the multi-gaussian for each class*/

                    Fold = 0.0;
                    for(k=0; k<dimStd[3]; k++)
                       {tmp[k] = -log(ls[k]+TINY) - 0.5*interEng[k] - extEng0[k*volSize];
                        Fold += qs[k] * (log(qs[k]+TINY) + tmp[k]);}

                    sqtmp = 0.0; /*NOTE: here NO tiny value added for normalization, following exact math*/
                    for(k=0; k<dimFull[3]; k++)
                       {qtmp[k] = l0[k*volSize] * exp(0.5 * interEng[lkp[k]] + extEng0[lkp[k]*volSize]);
                        sqtmp += qtmp[k];}  /*updating the posterior probability at the current voxel*/

                    for(k=0; k<dimFull[3]; k++)
                        q0[k*volSize] = qtmp[k]/sqtmp;  /*normalization and assign the updated posterior back to the posterior volume*/
/*NOTE: This approach is NOT ebdorsed by the MathWorks, because it changes the RHS argument (which points to const mxArray) and can lead to some strange side effects, but it does save memory a lot.*/

                    for(k=0; k<dimStd[3]; k++)
                        qs[k] = 0.0; /*initialize qs*/

                    for(k=0; k<dimFull[3]; k++)
                        qs[lkp[k]] += q0[k*volSize]; /*shrink the multi-gaussian for each class*/

                    Fnew = 0.0;
                    for(k=0; k<dimStd[3]; k++)
                       Fnew += qs[k] * (log(qs[k]+TINY) + tmp[k]);

                    dF += (Fnew - Fold); /*calculate the change in total free energy, which is used as a criterion for convergence of the algorithm*/
                }

             }

          }

     }
return dF;
}

static double updateLocal(float J[], double q[], double l[], double extEng[], int lkp[], mwSize dimStd[], mwSize dimFull[])
{
int it, k, volSize;
double Fold, Fnew, sqtmp, dF=0.0;
mwSize i0, i1, i2, i0start, i1start, i2start, offset, offsetNeighbor, offsetJ1;
double *q0, *l0, *extEng0;
double interEng[MAXCLASSES], ls[MAXCLASSES], qs[MAXCLASSES], qtmp[MAXCLASSES], tmp[MAXCLASSES];

volSize = dimStd[0]*dimStd[1]*dimStd[2];
offsetNeighbor = dimStd[3]*dimStd[3]*volSize;
offsetJ1=dimStd[3]*volSize;

for(it=0; it<2; it++)
    {
        i2start = it%2;
        for(i2=0; i2<dimStd[2]; i2++) /* Inferior -> Superior */
        {
            i1start = (i2start == (i2%2));
            for(i1=0; i1<dimStd[1]; i1++) /* Posterior -> Anterior */
            {
                i0start = (i1start == (i1%2));

                for(i0=i0start; i0<dimStd[0]; i0+=2) /* Left -> Right */
                {

                    offset = i0 + dimStd[0]*(i1 + dimStd[1]*i2);
/* NOTE: convert Matlab indexing into linear indexing, for COLUMN-first indexing (Matlab matrix style),
we have A(i0,i1,i2), is equivalent to A[i0 + dimStd[0]*(i1 + dimStd[1]*i2)], 0-based */

                    /* Pointers to current voxel in first volume */
                    q0 = q + offset;
                    l0 = l + offset;
                    extEng0 = extEng + offset;

                    for (k=0; k<dimStd[3]; k++)
                        { interEng[k] = 0.0;
                          ls[k] = 0.0;
                          qs[k] = 0.0; }  /*initialization*/

                    if (i2>0) /*inferior neighbor*/
                    neighborProcessLocal(interEng,J,q0-dimStd[0]*dimStd[1],lkp,dimStd,dimFull,volSize,offset,0*offsetNeighbor,offsetJ1);
                    if (i2<dimStd[2]-1) /*superior neighbor*/
                    neighborProcessLocal(interEng,J,q0+dimStd[0]*dimStd[1],lkp,dimStd,dimFull,volSize,offset,1*offsetNeighbor,offsetJ1);
                    if (i1>0) /*posterior neighbor*/
                    neighborProcessLocal(interEng,J,q0-dimStd[0],lkp,dimStd,dimFull,volSize,offset,2*offsetNeighbor,offsetJ1);
                    if (i1<dimStd[1]-1) /*anterior neighbor*/
                    neighborProcessLocal(interEng,J,q0+dimStd[0],lkp,dimStd,dimFull,volSize,offset,3*offsetNeighbor,offsetJ1);
                    if (i0>0) /*left neighbor*/
                    neighborProcessLocal(interEng,J,q0-1,lkp,dimStd,dimFull,volSize,offset,4*offsetNeighbor,offsetJ1);
                    if (i0<dimStd[0]-1) /*right neighbor*/
                    neighborProcessLocal(interEng,J,q0+1,lkp,dimStd,dimFull,volSize,offset,5*offsetNeighbor,offsetJ1);

                    for(k=0; k<dimFull[3]; k++)
                       { ls[lkp[k]] += l0[k*volSize];
                         qs[lkp[k]] += q0[k*volSize]; }  /*shrink the multi-gaussian for each class*/

                    Fold = 0.0;
                    for(k=0; k<dimStd[3]; k++)
                       {tmp[k] = -log(ls[k]+TINY) - 0.5*interEng[k] - extEng0[k*volSize];
                        Fold += qs[k] * (log(qs[k]+TINY) + tmp[k]);}

                    sqtmp = 0.0; /*NOTE: here NO tiny value added for normalization, following exact math*/
                    for(k=0; k<dimFull[3]; k++)
                       {qtmp[k] = l0[k*volSize] * exp(0.5 * interEng[lkp[k]] + extEng0[lkp[k]*volSize]);
                        sqtmp += qtmp[k];}  /*updating the posterior probability at the current voxel*/

                    for(k=0; k<dimFull[3]; k++)
                        q0[k*volSize] = qtmp[k]/sqtmp;  /*normalization and assign the updated posterior back to the posterior volume*/
/*NOTE: This approach is NOT ebdorsed by the MathWorks, because it changes the RHS argument (which points to const mxArray) and can lead to some strange side effects, but it does save memory a lot.*/

                    for(k=0; k<dimStd[3]; k++)
                        qs[k] = 0.0; /*initialize qs*/

                    for(k=0; k<dimFull[3]; k++)
                        qs[lkp[k]] += q0[k*volSize]; /*shrink the multi-gaussian for each class*/

                    Fnew = 0.0;
                    for(k=0; k<dimStd[3]; k++)
                       Fnew += qs[k] * (log(qs[k]+TINY) + tmp[k]);

                    dF += (Fnew - Fold); /*calculate the change in total free energy, which is used as a criterion for convergence of the algorithm*/
                }

             }

          }

     }
return dF;
}

static double updateRegional(float J[], double q[], double l[], double extEng[], int lkp[], float rLabel[], mwSize dimStd[], mwSize dimFull[])
{
int it, k, volSize;
double Fold, Fnew, sqtmp, dF=0.0;
mwSize i0, i1, i2, i0start, i1start, i2start, offset, offsetJ;
double *q0, *l0, *extEng0;
double interEng[MAXCLASSES], ls[MAXCLASSES], qs[MAXCLASSES], qtmp[MAXCLASSES], tmp[MAXCLASSES];

volSize = dimStd[0]*dimStd[1]*dimStd[2];

for(it=0; it<2; it++)
    {
        i2start = it%2;
        for(i2=0; i2<dimStd[2]; i2++) /* Inferior -> Superior */
        {
            i1start = (i2start == (i2%2));
            for(i1=0; i1<dimStd[1]; i1++) /* Posterior -> Anterior */
            {
                i0start = (i1start == (i1%2));

                for(i0=i0start; i0<dimStd[0]; i0+=2) /* Left -> Right */
                {

                    offset = i0 + dimStd[0]*(i1 + dimStd[1]*i2);
/* NOTE: convert Matlab indexing into linear indexing, for COLUMN-first indexing (Matlab matrix style),
we have A(i0,i1,i2), is equivalent to A[i0 + dimStd[0]*(i1 + dimStd[1]*i2)], 0-based */

                    /* Pointers to current voxel in first volume */
                    q0 = q + offset;
                    l0 = l + offset;
                    extEng0 = extEng + offset;

                    offsetJ = dimStd[3]*dimStd[3]*rLabel[offset]; /*which Global TCM to use, depending on regional labels*/

                    for (k=0; k<dimStd[3]; k++)
                        { interEng[k] = 0.0;
                          ls[k] = 0.0;
                          qs[k] = 0.0; }  /*initialization*/

                    if (i2>0) /*inferior neighbor*/
                    neighborProcessRegional(interEng,J,q0-dimStd[0]*dimStd[1],lkp,dimStd,dimFull,volSize,offsetJ);
                    if (i2<dimStd[2]-1) /*superior neighbor*/
                    neighborProcessRegional(interEng,J,q0+dimStd[0]*dimStd[1],lkp,dimStd,dimFull,volSize,offsetJ);
                    if (i1>0) /*posterior neighbor*/
                    neighborProcessRegional(interEng,J,q0-dimStd[0],lkp,dimStd,dimFull,volSize,offsetJ);
                    if (i1<dimStd[1]-1) /*anterior neighbor*/
                    neighborProcessRegional(interEng,J,q0+dimStd[0],lkp,dimStd,dimFull,volSize,offsetJ);
                    if (i0>0) /*left neighbor*/
                    neighborProcessRegional(interEng,J,q0-1,lkp,dimStd,dimFull,volSize,offsetJ);
                    if (i0<dimStd[0]-1) /*right neighbor*/
                    neighborProcessRegional(interEng,J,q0+1,lkp,dimStd,dimFull,volSize,offsetJ);

                    for(k=0; k<dimFull[3]; k++)
                       { ls[lkp[k]] += l0[k*volSize];
                         qs[lkp[k]] += q0[k*volSize]; }  /*shrink the multi-gaussian for each class*/

                    Fold = 0.0;
                    for(k=0; k<dimStd[3]; k++)
                       {tmp[k] = -log(ls[k]+TINY) - 0.5*interEng[k] - extEng0[k*volSize];
                        Fold += qs[k] * (log(qs[k]+TINY) + tmp[k]);}

                    sqtmp = 0.0; /*NOTE: here NO tiny value added for normalization, following exact math*/
                    for(k=0; k<dimFull[3]; k++)
                       {qtmp[k] = l0[k*volSize] * exp(0.5 * interEng[lkp[k]] + extEng0[lkp[k]*volSize]);
                        sqtmp += qtmp[k];}  /*updating the posterior probability at the current voxel*/

                    for(k=0; k<dimFull[3]; k++)
                        q0[k*volSize] = qtmp[k]/sqtmp;  /*normalization and assign the updated posterior back to the posterior volume*/
/*NOTE: This approach is NOT ebdorsed by the MathWorks, because it changes the RHS argument (which points to const mxArray) and can lead to some strange side effects, but it does save memory a lot.*/

                    for(k=0; k<dimStd[3]; k++)
                        qs[k] = 0.0; /*initialize qs*/

                    for(k=0; k<dimFull[3]; k++)
                        qs[lkp[k]] += q0[k*volSize]; /*shrink the multi-gaussian for each class*/

                    Fnew = 0.0;
                    for(k=0; k<dimStd[3]; k++)
                       Fnew += qs[k] * (log(qs[k]+TINY) + tmp[k]);

                    dF += (Fnew - Fold); /*calculate the change in total free energy, which is used as a criterion for convergence of the algorithm*/
                }

             }

          }

     }
return dF;
}

void mexFunction(int nlhs,mxArray *plhs[],int nrhs,const mxArray *prhs[])
{
    mwSize dimFull[4], dimStd[4]; /*dimFull: for posterior matrix including multi-gaussian; dimStd: for posterior matrix, one gaussian for each class*/
    mwSize i, N;
    int code;
    int *lkp;
    float *J;
    double *q, *l, *extEng, *deltaF;

    if (nrhs < 5 || nrhs >6 || nlhs>1)
        mexErrMsgTxt("Incorrect usage.");

    if (!mxIsNumeric(prhs[0]) || mxIsComplex(prhs[0]) || mxIsSparse(prhs[0]) || !mxIsSingle(prhs[0]))
        mexErrMsgTxt("TCM J must be numeric, real, full and single.");

    if (!mxIsNumeric(prhs[1]) || mxIsComplex(prhs[1]) || mxIsSparse(prhs[1]) )
        mexErrMsgTxt("Posterior probabilities must be numeric, real and full.");

    if (!mxIsNumeric(prhs[2]) || mxIsComplex(prhs[2]) || mxIsSparse(prhs[2]))
        mexErrMsgTxt("Likelihoods must be numeric, real and full.");

    if (!mxIsNumeric(prhs[3]) || mxIsComplex(prhs[3]) || mxIsSparse(prhs[3]))
        mexErrMsgTxt("External energy must be numeric, real and full.");

    if (!mxIsNumeric(prhs[4]) || mxIsComplex(prhs[4]) || mxIsSparse(prhs[4]) || !mxIsInt32(prhs[4]))
        mexErrMsgTxt("Multi-gaussian index must be numeric, real, full and integer (32 bit).");

    for(i=0; i<mxGetNumberOfDimensions(prhs[1]); i++)
        dimFull[i] = mxGetDimensions(prhs[1])[i];

    for(i=0; i<mxGetNumberOfDimensions(prhs[3]); i++)
        dimStd[i] = mxGetDimensions(prhs[3])[i];

    if (dimFull[3] > MAXCLASSES || dimStd[3] > MAXCLASSES)
        mexErrMsgTxt("Exceeds maximal tissue classes allowed.");

    if (dimFull[0] != dimStd[0] || dimFull[1] != dimStd[1] || dimFull[2] != dimStd[2])
        mexErrMsgTxt("Posterior probabilities and external energy have incompatible dimensions.");

    if (mxGetNumberOfDimensions(prhs[0])==2)
        {
         code = 0; /*Global TCM*/
         if (mxGetDimensions(prhs[0])[0] != dimStd[3] || mxGetDimensions(prhs[0])[1] != dimStd[3])
             mexErrMsgTxt("TCM J has incompatible dimensions.");
        }
    else if (mxGetNumberOfDimensions(prhs[0])==6) /*local TCM, dim1 x dim2 x dim3 x K x K x 6 matrix*/
        {
         code = 1; /*Local TCM, needs huge memory*/
         if (mxGetDimensions(prhs[0])[0] != dimStd[0] || mxGetDimensions(prhs[0])[1] != dimStd[1] || mxGetDimensions(prhs[0])[2] != dimStd[2] || mxGetDimensions(prhs[0])[3] != dimStd[3] || mxGetDimensions(prhs[0])[4] != dimStd[3] || mxGetDimensions(prhs[0])[5] != 6)
            mexErrMsgTxt("TCM J has incompatible dimensions.");
        }
    else if (mxGetNumberOfDimensions(prhs[0])==3)
        {
         code = 2; /*Regional TCM*/
         if (mxGetDimensions(prhs[0])[0] != dimStd[3] || mxGetDimensions(prhs[0])[1] != dimStd[3] || mxGetDimensions(prhs[0])[2] != 2)
             mexErrMsgTxt("TCM J has incompatible dimensions.");
         if (!mxIsNumeric(prhs[5]) || mxIsComplex(prhs[5]) || mxIsSparse(prhs[5]) || !mxIsSingle(prhs[5]))
             mexErrMsgTxt("Regional labels must be numeric, real, full and single.");
         if (mxGetDimensions(prhs[5])[0] != dimStd[0] || mxGetDimensions(prhs[5])[1] != dimStd[1] || mxGetDimensions(prhs[5])[2] != dimStd[2])
             mexErrMsgTxt("Regional labels have incompatible dimensions.");
        }
    else
        mexErrMsgTxt("TCM J has incompatible dimensions.");

    if (mxGetDimensions(prhs[2])[0] != dimFull[0] || mxGetDimensions(prhs[2])[1] != dimFull[1] || mxGetDimensions(prhs[2])[2] != dimFull[2] || mxGetDimensions(prhs[2])[3] != dimFull[3])
        mexErrMsgTxt("Likelihoods has incompatible dimensions.");

    if (mxGetNumberOfElements(prhs[4]) != dimFull[3])
        mexErrMsgTxt("Multi-gaussian index has incompatible dimensions.");

    J = (float *)mxGetData(prhs[0]);
    q = mxGetPr(prhs[1]);
    l = mxGetPr(prhs[2]);
    extEng = mxGetPr(prhs[3]);
    lkp = (int *)mxGetData(prhs[4]);

/*    for (i=0; i<dimFull[3]; i++) mexPrintf("%d   ",lkp[i]);
    mexPrintf("%d\n",sizeof(int));*/

    if ((lkp[dimFull[3]-1]+1) != dimStd[3])
       mexErrMsgTxt("Multi-gaussian index is incorrect.");

    /*GET DATA: MATRIX-->1D ARRAY, AND ARRAY STORED IN COLUMN-FIRST STYLE*/

    plhs[0] = mxCreateDoubleScalar(0.0);
    deltaF = mxGetPr(plhs[0]); /*CREATE OUTPUT*/

    /*mexPrintf("%f\n",*FEF);*/

    if (code == 0) /*Global TCM*/
       *deltaF = updateGlobal(J,q,l,extEng,lkp,dimStd,dimFull); /*posterior q will be updated in this function, since q is from prhs, which points to const mxArray, we modify *q and this is 
                                                                         NOT ebdorsed by the MathWorks, but it does save memory a lot*/

    if (code == 1) /*Local TCM*/
       *deltaF = updateLocal(J,q,l,extEng,lkp,dimStd,dimFull); /*posterior q will be updated in this function, since q is from prhs, which points to const mxArray, we modify *q and this is 
                                                                         NOT ebdorsed by the MathWorks, but it does save memory a lot*/

    if (code == 2) /*Regional TCM*/
       {
        float *rLabel;
        rLabel = (float *)mxGetData(prhs[5]);
        *deltaF = updateRegional(J,q,l,extEng,lkp,rLabel,dimStd,dimFull); /*posterior q will be updated in this function, since q is from prhs, which points to const mxArray, we modify *q and this is 
                                                                         NOT ebdorsed by the MathWorks, but it does save memory a lot*/
       }
    /*mexPrintf("%f\n",*deltaF);*/
}